### This file is prepared for JCGS with R package bst, version 0.3-14
### simulation study for example 4 with quadratic majorization for robust C-loss (ClossBoostQM)
### ref: Quadratic Majorization for Nonconvex Loss with Applications to the Boosting Algorithm, Zhu Wang, 2017 (the paper).

### run as-is to generate results

### for multi-class problems
library("mvtnorm")
library("bst")
nrun <- 50
ex <- 4
J <- 3
if(J < 3) stop("this is for multiclass problem\n")
if(ex==8) 
    p <- 0
if(ex!=8) 
    p <- 18
rf <- "closs" ### rf can be "thinge" which also provides results for HingeBoost2
tuning <- "error"
twin <- FALSE
rho <- 0.0
if(ex ==4) 
    learner <- "ls"
if(ex==8) 
    learner <- "sm"
mstop1 <- mstop2 <- optqu <- rep(NA, nrun)
summary7 <- function(x, na.rm=TRUE) c(summary(x, na.rm=na.rm), sd=sd(x, na.rm=na.rm))
### random data generating
datagen <- function(ntr, ntu, nte, ex, percon, rf, p=18, sigma=NULL, k){
    if(ex %in% c(4, 8)){
        n <- ntr+ntu+nte
        r <- runif(n, 0, 1)
        theta <- runif(n, 0, 2*pi)
        x1 <- sqrt(r)*cos(theta)
        x2 <- sqrt(r)*sin(theta)
        if(ex==4)
            y <- as.integer(k*theta/(2*pi))+1    ### Example 4.1 in Wu and Liu (2007), JASA
	else {
            y1 <- as.integer(k*theta/(pi))+1    ### Example 4.2 in Wu and Liu (2007), JASA. Note, the form is corrected from 2pi to pi here
            y <- rep(NA, length(y1))
            y[y1 %in% c(1, 4)] <- 1
            y[y1 %in% c(2, 3)] <- 2
            y[y1 %in% c(5, 6)] <- 3
        }
	xnois <- matrix(runif(n*p, -1, 1), ncol=p)
        if(p > 0)
            x <- cbind(x1, x2, xnois) 
        else x <- cbind(x1, x2)
        xnonzero <- 1:2
        if(p > 0)
            xzero <- 3:(p+2)
        else xzero <- 0
        d <- p+2
        b <- NA
        A <- NA
    }
    x <- as.data.frame(x)
### contaminated data
    xtr <- x[1:ntr,]
    ytr <- y[1:ntr]
    xtu <- x[(1+ntr):(ntr+ntu),]
    ytu <- y[(1+ntr):(ntr+ntu)]
    xte <- x[(1+ntr+ntu):n,]
    yte <- y[(1+ntr+ntu):n]
    con <- sample(ntr) 
    contu <- sample(ntu) 
    conte <- sample(nte) 
    ytrold <- ytr
    ytuold <- ytu
    yteold <- yte
    if(percon > 0){
        j <- con[1:(percon*ntr)]
        jtu <- contu[1:(percon*ntu)]
        jte <- conte[1:(percon*nte)]
        for(jj in j){
            tmp <- ytr[jj] 
            ytr[jj] <- sample(c(1:J)[-tmp], size=1)
	}
        for(jj in jtu){
            tmp <- ytu[jj] 
            ytu[jj] <- sample(c(1:J)[-tmp], size=1)
	}
        for(jj in jte){
            tmp <- yte[jj] 
            yte[jj] <- sample(c(1:J)[-tmp], size=1)
	}
    }
                                        #xtu <- xte; ytu <- ytuold <- yte
                                        #xte <- xtr; yte <- yteold <- ytr
    RET <- list(xtr=xtr, ytr=ytr, xtu=xtu, ytu=ytu, xte=xte, yte=yte, ytrold=ytrold, ytuold=ytuold, yteold=yteold, d=d, xnonzero=xnonzero, xzero=xzero, b=b, A=A)
}
ntr <- 100
ntu <- 1000 
nte <- 10000
n <- ntr+ntu+nte
bsttype1 <- c("ClossBoost", "HingeBoost")
rbsttype1 <- c("ClossBoostQM", "THingeBoost")
if(rf=="closs") {rfk <- 1; ms <- 200
}else if(rf=="thinge") {
    rfk <- 2; ms <- 400
}
per <- c(0, 5, 10, 20)/100
for(k in (1:length(per))[4:1]){
    nmethod <- 3
    err <- errb <- errold <- sen <- spc <- matrix(NA, ncol=nmethod, nrow=nrun)
    colnames(err) <- colnames(errold) <- colnames(sen) <- colnames(spc) <- c(paste(bsttype1[rfk], "(clean data)"), bsttype1[rfk], rbsttype1[rfk])
    nvar.m <- nvar.m2 <- err.m <- err.m2 <- matrix(NA, ncol=ms, nrow=nrun)
    for(i in 1:nrun){
        cat("i=", i, "\n")
        set.seed(1000+i)
### generate unifrom distribution over a unit disk, cf: Wu and Liu (2007), Park and Liu (2011), http://mathworld.wolfram.com/DiskPointPicking.html
        dat <- datagen(ntr, ntu, nte, ex=ex, percon=per[k], p=p, rf=rf, k=J)
        xzero <- dat$xzero
        xnonzero <- dat$xnonzero
        xtr <- dat$xtr
        xtu <- dat$xtu
        xte <- dat$xte
        ytr <- dat$ytr
        ytu <- dat$ytu
        yte <- dat$yte
        ytrold <- dat$ytrold
        ytuold <- dat$ytuold
        yteold <- dat$yteold
        b <- dat$b
        d <- dat$d
        A <- dat$A
### results from QMBA robust boosting
	if(rf=="thinge") bstfam <- "hinge2"
	else if(rf=="closs") bstfam <- "closs"
        if(rf=="thinge"){
            s <- seq(0, 2, by=0.5)
        }
	else if(rf=="closs") s <- seq(0.5, 2, by=0.5)
	mstop <- ms
        err1 <- rep(NA, length(s))
###outloop on s
        fit <- vector("list", length(s))
        for(ks in 1:length(s)){
            if(rf %in% c("thinge", "closs")){
		fit[[ks]] <- rmbst(xtr, ytr, ctrl = bst_control(s=s[ks], mstop=mstop, trace=FALSE, center=FALSE, iter=100, threshold="standard"), rfamily = rf, learner = learner, del=1e-20)
                                        #   if(twin)
                                        #    fit[[ks]] <- rmbst(xtr, ytr, ctrl = bst_control(s=s[ks], mstop=mstop, twinboost=TRUE, center=FALSE, iter=100, f.init=predict(fit[[ks]]), xselect.init = fit[[ks]]$xselect),rfamily = rf, learner = learner, del=1e-16)
            }
            res2 <- res3 <- rep(NA, length(mstop))
                                        # for(jk in 1:length(mstop)){
                                        #res1 <- predict(fit[[ks]], newdata=xtu, newy=ytu, type=tuning)  ### may delete this and next line
                                        # res2 <- min(res1)
                                        #res2[jk] <- min(res1)
                                        #}
### for fixed s, find the best model
                                        #dat.m2 <- fit[[ks]][[which.min(res2)]]
            dat.m2 <- fit[[ks]]
            res1 <- predict(dat.m2, newdata=xtu, newy=ytu, type=tuning)
            res <- predict(dat.m2, newdata=xte, newy=yte, type="error")
            resold <- predict(dat.m2, newdata=xte, newy=yteold, type=tuning)
                                        #if(p <= 2) minm <- mstop[length(mstop)]
                                        #if(p <= 2) minm <- dat.m2$ctrl$mstop
                                        #else
	    minm <- which.min(res1)
                                        #    if(p < ntr)
                                        #	    minm <- max(40, which.min(res1))
                                        #    else minm <- max(10, which.min(res1))
            err1[ks] <- res1[minm]
        }
###find smallest error among different s
        findmin <- which.min(err1)
        dat.m2 <- fit[[findmin]]
	if(twin)
            dat.m2 <- rmbst(xtr, ytr, ctrl = bst_control(s=s[ks], mstop=mstop, twinboost=TRUE, center=FALSE, iter=100, f.init=predict(dat.m2), xselect.init = dat.m2$xselect),rfamily = rf, learner = learner, del=1e-16)
### for the best s, find the best model (or best mstop)
                                        #    res2 <- rep(NA, length(mstop))
                                        #    for(jk in 1:length(mstop)){
                                        #res1 <- predict(dat.m2[[jk]], newdata=xtu, newy=ytu, type=tuning)
        res1 <- predict(dat.m2, newdata=xtu, newy=ytu, type=tuning)
	res2 <- min(res1)
                                        #res2[jk] <- res1[length(res1)] ### changed 3/28/15
					#        res2[jk] <- min(res1)
                                        #    }
        optqu[i] <- findmin
        cat("\n which s has smallest error/loss", s[findmin], "\n")
                                        #cat("dat.m$yhat - dat.m2$yhat", summary(dat.m$yhat - dat.m2$yhat), "\n")
###find smallest error among different mstop
                                        #    dat.m2 <- dat.m2[[which.min(res2)]]
        res1 <- predict(dat.m2, newdata=xtu, newy=ytu, type=tuning)

        res <- predict(dat.m2, newdata=xte, newy=yte, type="error")
	err.m2[i,1:dat.m2$ctrl$mstop] <- res
        for(jjj in 1: dat.m2$ctrl$mstop)
                                        #nvar.m2[i,jjj] <- length(unique(unlist(dat.m2$ensemble[1:jjj])))
            nvar.m2[i,jjj] <- sum(!is.na(unique(unlist(dat.m2$ensemble[1:jjj]))))
        mstop1[i] <- mstop[which.min(res2)]
                                        #if(p <= 2) minm <- dat.m2$ctrl$mstop
                                        #                                else
                                        #    if(p < ntr)
        minm <- which.min(res1)
					#if(ex==8 && twin)
					#minm <- max(10, which.min(res1))
                                        #    else minm <- max(10, which.min(res1))
                                        #minm <- dat.m2$ctrl$mstop    ### changed 3/28/15
        mstop2[i] <- minm
        err[i,3] <- res[minm]
        errb[i,3] <- min(res)  ### best value in the test data
                                        #plot(res, col="red")
                                        #points(predict(dat.m, newdata=xte, newy=yte, type="error"))
                                        #browser()
        errold[i,3] <- resold[minm]
        sen[i,3] <- sum(xnonzero %in% unique(unlist(dat.m2$ensemble[1:minm])))/length(xnonzero)
        spc[i,3] <- sum(!xzero %in% unique(unlist(dat.m2$ensemble[1:minm])))/length(xzero)
                                        #sen[i,3] <- sum(c(1:2) %in% unique(dat.m2$ensemble[1:minm]))/2
                                        #spc[i,3] <- sum(!c(3:(p+2)) %in% unique(dat.m2$ensemble[1:minm]))/p
                                        #relative prediction error, see Zou JASA paper above
                                        #if(ex > 2) errold <- errold/dat$sigma2
                                        #}
###results from boosting
        dat.m0 <- mbst(xtr, ytrold, ctrl = bst_control(mstop=ms, center=FALSE, trace=FALSE, s=s[findmin]), family = bstfam, learner = learner)
	res1 <- predict(dat.m0, newdata=xtu, newy=ytuold, type=tuning)
	res <- predict(dat.m0, newdata=xte, newy=yteold, type="error")
        resold <- predict(dat.m0, newdata=xte, newy=yteold, type=tuning)
        err[i,1] <- res[which.min(res1)]
        errold[i,1] <- resold[which.min(res1)]
	sen[i,1] <- sum(xnonzero %in% unique(unlist(dat.m0$ensemble[1:which.min(res1)])))/length(xnonzero)
        spc[i,1] <- sum(!xzero %in% unique(unlist(dat.m0$ensemble[1:which.min(res1)])))/length(xzero)
                                        #bstfam <- "hinge"
	if(bstfam %in% c("hinge2", "closs")){
            dat.m <- mbst(xtr, ytr, ctrl = bst_control(mstop=ms, center=FALSE, trace=FALSE, s=s[findmin]), family = bstfam, learner = learner)
            if(twin)
                dat.m <- mbst(xtr, ytr, ctrl = bst_control(mstop=ms, s=s[findmin], twinboost=TRUE, center=FALSE, f.init=predict(dat.m), xselect.init = dat.m$xselect), family = bstfam, learner = learner)
        }
        else dat.m <- mhingebst(xtr, ytr, ctrl = bst_control(mstop=ms, center=FALSE, trace=FALSE), family = bstfam, learner = learner)
        res1 <- predict(dat.m, newdata=xtu, newy=ytu, type=tuning)
        res <- predict(dat.m, newdata=xte, newy=yte, type="error")
        resold <- predict(dat.m, newdata=xte, newy=yteold, type=tuning)
        err[i,2] <- res[which.min(res1)]
        errb[i,2] <- min(res)  ### best value in the test data
        if(ex %in% c(4, 8))
            err.m[i,1:dat.m$ctrl$mstop] <- res
        errold[i,2] <- resold[which.min(res1)]
        sen[i,2] <- sum(xnonzero %in% unique(unlist(dat.m$ensemble[1:which.min(res1)])))/length(xnonzero)
        spc[i,2] <- sum(!xzero %in% unique(unlist(dat.m$ensemble[1:which.min(res1)])))/length(xzero)
	for(jjj in 1: dat.m$ctrl$mstop)
            nvar.m[i,jjj] <- sum(!is.na(unique(unlist(dat.m$ensemble[1:jjj]))))
        if(i %% 1==0){
            cat("percentage of contamination", per[k], "\n")
            cat("i=", i, "\n")
            cat("s selected\n")
            print(table(s[optqu]))
                                        #cat("optimal mstop selected by tuning data in different M\n")
                                        #print(summary(mstop1))
                                        #cat("mstop selected by tuning data\n")
                                        #print(summary(mstop2))
            cat("test error for contaminated test data\n")
            print(apply(err, 2, summary7, na.rm=TRUE))
                                        #}   
					#cat("sensitivity\n")
					#print(apply(sen, 2, summary, na.rm=TRUE))
					#cat("specificity\n")
					#print(apply(spc, 2, summary, na.rm=TRUE))
            cat("number of variables selected\n")
            print(apply((1-spc)*length(xzero)+sen*length(xnonzero), 2, summary7, na.rm=TRUE))
        }
        nvar <- (1-spc)*length(xzero)+sen*length(xnonzero)
        res <- list(contscen=k, p=d, rf=rf, bstfam=bstfam, s=s, err=err, errb=errb, errold=errold, sen=sen, spc=spc, mstop2=mstop2, optqu=optqu, nvar=nvar, err.m=err.m, err.m2=err.m2, nvar.m=nvar.m, nvar.m2=nvar.m2)
        if(i == nrun){
                                        #if(ex %in% c(1, 2))
            pdf(file=paste("mclerr", rf, ".pdf", sep=""))
            par(mfrow=c(2, 1))
                                        #        if(k==4 && ex==6 && ntr==200)
                                        #        ylim1 <- c(0.27, 0.47)
                                        #        else ylim1 <- NULL
            if(ms==100){
                boxplot(cbind(res$err.m[, 25], res$err.m2[,25], res$err.m[, 50], res$err.m2[,50], res$err.m[,75], res$err.m2[, 75], res$err.m[, 100], res$err.m2[, 100]), names=c("25", "25", "50", "50", "75", "75", "100", "100"), col=rep(c("blue", "red"), 4), lty=rep(c(1, 2), 4), ylab="Misclassification Error")
                boxplot(cbind(res$nvar.m[, 25], res$nvar.m2[,25], res$nvar.m[, 50], res$nvar.m2[,50], res$nvar.m[, 75], res$nvar.m2[, 75], res$nvar.m[, 100], res$nvar.m2[, 100]),
                        names=c("25", "25", "50", "50", "75", "75", "100", "100"), col=rep(c("blue" ,"red"), 4), lty=rep(c(1, 2), 4), ylab="No. Variables")
            }
            if(ms==200){
                boxplot(cbind(res$err.m[, 50], res$err.m2[,50], res$err.m[, 100], res$err.m2[,100], res$err.m[, 150], res$err.m2[, 150], res$err.m[,200], res$err.m2[, 200]), names=c("B50", "TB50", "B100", "TB100", "B150", "TB150", "B200", "TB200"), col=rep(c("red", "blue"), 4), lty=rep(c(1, 2), 4), ylab="Misclassification Error")
                                        #        boxplot(cbind(res$err.m[, 20], res$err.m2[,20], res$err.m[, 40], res$err.m2[,40], res$err.m[, 60], res$err.m2[, 60], res$err.m[, 80], res$err.m2[, 80], res$err.m[, 100], res$err.m2[, 100]), names=c("B20", "TB20", "B40", "TB40", "B60", "TB60", "B80", "TB80", "B100", "TB100"), col=rep(c("red", "blue"), 5), lty=rep(c(1, 2), 5), ylab="Misclassification Error")
                                        #        boxplot(cbind(res$nvar.m[, 20], res$nvar.m2[,20], res$nvar.m[, 40], res$nvar.m2[,40], res$nvar.m[, 60], res$nvar.m2[, 60], res$nvar.m[, 80], res$nvar.m2[, 80], res$nvar.m[, 100], res$nvar.m2[, 100]), names=c("B20", "TB20", "B40", "TB40", "B60", "TB60", "B80", "TB80", "B100", "TB100"), col=rep(c("red", "blue"), 5), lty=rep(c(1, 2), 5), ylab="No. Variables")
                boxplot(cbind(res$nvar.m[, 50], res$nvar.m2[,50], res$nvar.m[, 100], res$nvar.m2[,100], res$nvar.m[, 150], res$nvar.m2[, 150], res$nvar.m[, 200], res$nvar.m2[, 200]), names=c("B50", "TB50", "B100", "TB100", "B150", "TB150", "B200", "TB200"), col=rep(c("red", "blue"), 4), lty=rep(c(1, 2), 4), ylab="No. Variables")
            }
            if(ms==300){
                boxplot(cbind(res$err.m[, 50], res$err.m2[,50], res$err.m[, 100], res$err.m2[,100], res$err.m[, 150], res$err.m2[, 150], res$err.m[,200], res$err.m2[, 200], res$err.m[,250], res$err.m2[,250], res$err.m[,300], res$err.m2[,300]), names=c("50", "50", "100", "100", "150", "150", "200", "200","250", "250", "300", "300"), col=rep(c("red", "blue"), 4), lty=rep(c(1, 2), 4), ylab="Misclassification Error")
                boxplot(cbind(res$nvar.m[, 50], res$nvar.m2[,50], res$nvar.m[, 100], res$nvar.m2[,100], res$nvar.m[, 150], res$nvar.m2[, 150], res$nvar.m[, 200], res$nvar.m2[, 200], res$nvar.m[,250], res$nvar.m2[,250], res$nvar.m[,300], res$nvar.m2[,300]), names=c("50", "50", "100", "100", "150", "150", "200", "200", "250", "250", "300", "300"), col=rep(c("red", "blue"), 4), lty=rep(c(1, 2), 4), ylab="No. Variables")
	    } 
            else if(ms==400){
                boxplot(cbind(res$err.m[, 100], res$err.m2[,100], res$err.m[,200], res$err.m2[, 200], res$err.m[, 300], res$err.m2[,300], res$err.m[, 400], res$err.m2[, 400]), col=rep(c("red", "blue"), 4), lty=rep(c(1, 2), 4),
                        names=c("B100", "TB100", "B200", "TB200", "B300", "TB300", "B400", "TB400"), ylab="Misclassification Error")
                                        #names=c("B50", "TB50", "B100", "TB100", "B150", "TB150", "B200", "TB200"), col=rep(c("red", "blue"), 4), lty=rep(c(1, 2), 4), ylab="Misclassification Error")
                boxplot(cbind(res$nvar.m[, 100], res$nvar.m2[,100], res$nvar.m[, 200], res$nvar.m2[,200], res$nvar.m[, 300], res$nvar.m2[, 300], res$nvar.m[, 400], res$nvar.m2[, 400]), 
                        names=c("B100", "TB100", "B200", "TB200", "B300", "TB300", "B400", "TB400"),
                        col=rep(c("red", "blue"), 4), lty=rep(c(1, 2), 4),
                        ylab="No. Variables")
            }
            else if(ms==600){
                boxplot(cbind(res$err.m[, 100], res$err.m2[,100], res$err.m[,200], res$err.m2[, 200], res$err.m[, 300], res$err.m2[,300], res$err.m[, 400], res$err.m2[, 400], res$err.m[,500], res$err.m2[,500], res$err.m[,600], res$err.m2[,600]), col=rep(c("red", "blue"), 6), lty=rep(c(1, 2), 6),
                        names=c("100", "100", "200", "200", "300", "300", "400", "400", "500", "500", "600", "600"), ylab="Misclassification Error")
                                        #names=c("B50", "TB50", "B100", "TB100", "B150", "TB150", "B200", "TB200"), col=rep(c("red", "blue"), 4), lty=rep(c(1, 2), 4), ylab="Misclassification Error")
                boxplot(cbind(res$nvar.m[, 100], res$nvar.m2[,100], res$nvar.m[, 200], res$nvar.m2[,200], res$nvar.m[, 300], res$nvar.m2[, 300], res$nvar.m[, 400], res$nvar.m2[, 400], res$nvar.m[,500], res$nvar.m2[,500], res$nvar.m[,600], res$nvar.m2[,600]), 
                        names=c("100", "100", "200", "200", "300", "300", "400", "400", "500", "500", "600", "600"), 
                        col=rep(c("red", "blue"), 6), lty=rep(c(1, 2), 6),
                        ylab="No. Variables")
            }
            dev.off()
        }
    }
    if(i==nrun){
        filename <- paste("ex", ex, "p", d, "con", k,"_rf", rf, "_bstfam", bstfam, "_twin", twin, "_n", ntr, "_rho",rho, learner, substr(dat.m2$ctrl$type, 1, 4), ".Rda", sep="")
        cat("saved RDa file name: ", filename, "\n")
                                        #save(file=filename, res)
    }
}
sessionInfo();
q("no")

