### This file is prepared for JCGS with R package bst, version 0.3-14
### simulation study for example 4 with multi-class HingeBoost, HingeBoost2 and regularized multinomial algorithms
### ref: Quadratic Majorization for Nonconvex Loss with Applications to the Boosting Algorithm, Zhu Wang, 2017 (the paper).

### user configuration
bstfam <- "hinge2"
### bstfam can be either one of the following: "hinge", "hinge2", "multinomial"
### and corresponding to "HingeBoost2", "HingeBoost" (NOT mistakes following the R package bst) and "Multinomial" in the paper, see Table 9 and 10
### end of configuration

bstfamname <- switch(bstfam,
	"hinge"="HingeBoost2",
	"hinge2"="HingeBoost",
	"multinomial"="Multinomial")
### for multi-class problems
library("mvtnorm")
library("bst")
library("glmnet")
nrun <- 50
ex <- 4
J <- 3
if(J < 3) stop("this is for multiclass problem\n")
if(ex==8) 
    p <- 0
if(ex!=8) 
    p <- 18
rf <- "closs"
tuning <- "error"
twin <- FALSE
rho <- 0.0
if(ex ==4) 
    learner <- "ls"
if(ex==8) 
    learner <- "sm"
mstop1 <- mstop2 <- optqu <- rep(NA, nrun)
summary7 <- function(x, na.rm=TRUE) c(summary(x, na.rm=na.rm), sd=sd(x, na.rm=na.rm))
### random data generating
datagen <- function(ntr, ntu, nte, ex, percon, rf, p=18, sigma=NULL, k){
    if(ex %in% c(4, 8)){
        n <- ntr+ntu+nte
        r <- runif(n, 0, 1)
        theta <- runif(n, 0, 2*pi)
        x1 <- sqrt(r)*cos(theta)
        x2 <- sqrt(r)*sin(theta)
        if(ex==4)
            y <- as.integer(k*theta/(2*pi))+1    ### Example 4.1 in Wu and Liu (2007), JASA
	else {
            y1 <- as.integer(k*theta/(pi))+1    ### Example 4.2 in Wu and Liu (2007), JASA. Note, the form is corrected from 2pi to pi here
            y <- rep(NA, length(y1))
            y[y1 %in% c(1, 4)] <- 1
            y[y1 %in% c(2, 3)] <- 2
            y[y1 %in% c(5, 6)] <- 3
        }
	xnois <- matrix(runif(n*p, -1, 1), ncol=p)
        if(p > 0)
            x <- cbind(x1, x2, xnois) 
        else x <- cbind(x1, x2)
        xnonzero <- 1:2
        if(p > 0)
            xzero <- 3:(p+2)
        else xzero <- 0
        d <- p+2
        b <- NA
        A <- NA
    }
    x <- as.data.frame(x)
### contaminated data
    xtr <- x[1:ntr,]
    ytr <- y[1:ntr]
    xtu <- x[(1+ntr):(ntr+ntu),]
    ytu <- y[(1+ntr):(ntr+ntu)]
    xte <- x[(1+ntr+ntu):n,]
    yte <- y[(1+ntr+ntu):n]
    con <- sample(ntr) 
    contu <- sample(ntu) 
    conte <- sample(nte) 
    ytrold <- ytr
    ytuold <- ytu
    yteold <- yte
    if(percon > 0){
        j <- con[1:(percon*ntr)]
        jtu <- contu[1:(percon*ntu)]
        jte <- conte[1:(percon*nte)]
        for(jj in j){
            tmp <- ytr[jj] 
            ytr[jj] <- sample(c(1:J)[-tmp], size=1)
	}
        for(jj in jtu){
            tmp <- ytu[jj] 
            ytu[jj] <- sample(c(1:J)[-tmp], size=1)
	}
        for(jj in jte){
            tmp <- yte[jj] 
            yte[jj] <- sample(c(1:J)[-tmp], size=1)
	}
    }
                                        #xtu <- xte; ytu <- ytuold <- yte
                                        #xte <- xtr; yte <- yteold <- ytr
    RET <- list(xtr=xtr, ytr=ytr, xtu=xtu, ytu=ytu, xte=xte, yte=yte, ytrold=ytrold, ytuold=ytuold, yteold=yteold, d=d, xnonzero=xnonzero, xzero=xzero, b=b, A=A)
}
ntr <- 100
ntu <- 1000 
nte <- 10000
n <- ntr+ntu+nte
ms <- 200
if(bstfam=="hinge2") ms <- 400
per <- c(0, 5, 10, 20)/100
for(k in (1:length(per))[4:1]){
    nmethod <- 2
    nvar <- err <- errb <- errold <- sen <- spc <- matrix(NA, ncol=nmethod, nrow=nrun)
    colnames(err) <- colnames(errold) <- colnames(sen) <- colnames(spc) <- colnames(nvar) <- c("None", bstfamname)
    nvar.m <- nvar.m2 <- err.m <- err.m2 <- matrix(NA, ncol=ms, nrow=nrun)
    for(i in 1:nrun){
	set.seed(1000+i)
### generate unifrom distribution over a unit disk, cf: Wu and Liu (2007), Park and Liu (2011), http://mathworld.wolfram.com/DiskPointPicking.html
        dat <- datagen(ntr, ntu, nte, ex=ex, percon=per[k], p=p, rf=rf, k=J)
        xzero <- dat$xzero
        xnonzero <- dat$xnonzero
        xtr <- dat$xtr
        xtu <- dat$xtu
        xte <- dat$xte
        ytr <- dat$ytr
        ytu <- dat$ytu
        yte <- dat$yte
        ytrold <- dat$ytrold
        ytuold <- dat$ytuold
        yteold <- dat$yteold
        b <- dat$b
        d <- dat$d
        A <- dat$A
### results from QMBA robust boosting
#if(rf=="thinge") bstfam <- "hinge2"
#	else if(rf=="closs") bstfam <- "closs"
#        if(rf=="thinge"){
	#            s <- seq(0, 2, by=0.5)
	#        }
	#	else if(rf=="closs") s <- seq(0.5, 2, by=0.5)
	mstop <- ms
###results from boosting
	if(bstfam %in% c("hinge", "hinge2")){
            if(bstfam %in% c("hinge2")){
                dat.m <- mbst(xtr, ytr, ctrl = bst_control(mstop=ms, center=FALSE, trace=FALSE), family = bstfam, learner = learner)
                if(twin)
                    dat.m <- mbst(xtr, ytr, ctrl = bst_control(mstop=ms, twinboost=TRUE, center=FALSE, f.init=predict(dat.m), xselect.init = dat.m$xselect), family = bstfam, learner = learner)
            }
            else if(bstfam=="hinge") 
                dat.m <- mhingebst(xtr, ytr, ctrl = bst_control(mstop=ms, center=FALSE, trace=FALSE), family = bstfam, learner = learner)       
            res1 <- predict(dat.m, newdata=xtu, newy=ytu, type=tuning)
            res <- predict(dat.m, newdata=xte, newy=yte, type="error")
            resold <- predict(dat.m, newdata=xte, newy=yteold, type=tuning)
	    err[i,2] <- res[which.min(res1)]
            errb[i,2] <- min(res)  ### best value in the test data
            if(ex %in% c(4, 8))
                err.m[i,1:dat.m$ctrl$mstop] <- res
            errold[i,2] <- resold[which.min(res1)]
            sen[i,2] <- sum(xnonzero %in% unique(unlist(dat.m$ensemble[1:which.min(res1)])))/length(xnonzero)
            spc[i,2] <- sum(!xzero %in% unique(unlist(dat.m$ensemble[1:which.min(res1)])))/length(xzero)
            for(jjj in 1: dat.m$ctrl$mstop)
                nvar.m[i,jjj] <- sum(!is.na(unique(unlist(dat.m$ensemble[1:jjj]))))        
        }
	else {
            dat.m <- glmnet(x=as.matrix(xtr), y=ytr, family="multinomial")
            tmp1 <- predict(dat.m, newx=as.matrix(xtu), type="class")
	    res1 <- apply(tmp1, 2, function(x) mean(x != ytu))
            tmp <- predict(dat.m, newx=as.matrix(xte), type="class")
	    res <- apply(tmp, 2, function(x) mean(x != yte))
            err[i,2] <- res[which.min(res1)]
            tmp <- NULL
            for(j in 1:J)
	    tmp <- c(tmp, predict(dat.m, type="nonzero")[[j]][which.min(res1)])
            tmp <- unique(unlist(tmp))
	    sen[i,2] <- sum(xnonzero %in% tmp)/length(xnonzero)
	    spc[i,2] <- sum(!xzero %in% tmp)/length(xzero)
            nvar[i,2] <- length(tmp)
    }
	
        if(i %% 1==0){
            cat("percentage of contamination", per[k], "\n")
            cat("bstfam=", bstfam, "\n")
            cat("i=", i, "\n")
	    cat("test error for contaminated test data\n")
	    #print(apply(err, 2, summary7, na.rm=TRUE))
            print(summary7(err[,2], na.rm=TRUE))
                                        #}   
                                        #else
                                        #    cat("test error for non-contaminated test data\n")
					#cat("sensitivity\n")
					#print(apply(sen, 2, summary, na.rm=TRUE))
					#cat("specificity\n")
					#print(apply(spc, 2, summary, na.rm=TRUE))
            cat("number of variables selected\n")
            nvar <- (1-spc)*length(xzero)+sen*length(xnonzero)
	    #if(bstfam %in% c("hinge", "hinge2"))
	    #print(apply((1-spc)*length(xzero)+sen*length(xnonzero), 2, summary, na.rm=TRUE))
	    print(summary7(nvar[,2], na.rm=TRUE))
	    #else print(apply(nvar, 2, summary7, na.rm=TRUE))               
    }
	if(bstfam %in% c("hinge", "hinge2"))
	#nvar <- (1-spc)*length(xzero)+sen*length(xnonzero)
        res <- list(contscen=k, p=d, rf=rf, bstfam=bstfam, err=err, errb=errb, errold=errold, sen=sen, spc=spc, mstop2=mstop2, optqu=optqu, nvar=nvar, err.m=err.m, err.m2=err.m2, nvar.m=nvar.m, nvar.m2=nvar.m2)
        if(i == nrun){
                                        #if(ex %in% c(1, 2))
            pdf(file=paste("mclerrNonrobust", rf, ".pdf", sep=""))
            par(mfrow=c(2, 1))
                                        #        if(k==4 && ex==6 && ntr==200)
                                        #        ylim1 <- c(0.27, 0.47)
                                        #        else ylim1 <- NULL
            if(ms==100){
                boxplot(cbind(res$err.m[, 25], res$err.m2[,25], res$err.m[, 50], res$err.m2[,50], res$err.m[,75], res$err.m2[, 75], res$err.m[, 100], res$err.m2[, 100]), names=c("25", "25", "50", "50", "75", "75", "100", "100"), col=rep(c("blue", "red"), 4), lty=rep(c(1, 2), 4), ylab="Misclassification Error")
                boxplot(cbind(res$nvar.m[, 25], res$nvar.m2[,25], res$nvar.m[, 50], res$nvar.m2[,50], res$nvar.m[, 75], res$nvar.m2[, 75], res$nvar.m[, 100], res$nvar.m2[, 100]),
                        names=c("25", "25", "50", "50", "75", "75", "100", "100"), col=rep(c("blue" ,"red"), 4), lty=rep(c(1, 2), 4), ylab="No. Variables")
            }
	    if(ms==200 && bstfam %in% c("hinge", "hinge2")){
                boxplot(cbind(res$err.m[, 50], res$err.m2[,50], res$err.m[, 100], res$err.m2[,100], res$err.m[, 150], res$err.m2[, 150], res$err.m[,200], res$err.m2[, 200]), names=c("B50", "TB50", "B100", "TB100", "B150", "TB150", "B200", "TB200"), col=rep(c("red", "blue"), 4), lty=rep(c(1, 2), 4), ylab="Misclassification Error")
                                        #        boxplot(cbind(res$err.m[, 20], res$err.m2[,20], res$err.m[, 40], res$err.m2[,40], res$err.m[, 60], res$err.m2[, 60], res$err.m[, 80], res$err.m2[, 80], res$err.m[, 100], res$err.m2[, 100]), names=c("B20", "TB20", "B40", "TB40", "B60", "TB60", "B80", "TB80", "B100", "TB100"), col=rep(c("red", "blue"), 5), lty=rep(c(1, 2), 5), ylab="Misclassification Error")
                                        #        boxplot(cbind(res$nvar.m[, 20], res$nvar.m2[,20], res$nvar.m[, 40], res$nvar.m2[,40], res$nvar.m[, 60], res$nvar.m2[, 60], res$nvar.m[, 80], res$nvar.m2[, 80], res$nvar.m[, 100], res$nvar.m2[, 100]), names=c("B20", "TB20", "B40", "TB40", "B60", "TB60", "B80", "TB80", "B100", "TB100"), col=rep(c("red", "blue"), 5), lty=rep(c(1, 2), 5), ylab="No. Variables")
                boxplot(cbind(res$nvar.m[, 50], res$nvar.m2[,50], res$nvar.m[, 100], res$nvar.m2[,100], res$nvar.m[, 150], res$nvar.m2[, 150], res$nvar.m[, 200], res$nvar.m2[, 200]), names=c("B50", "TB50", "B100", "TB100", "B150", "TB150", "B200", "TB200"), col=rep(c("red", "blue"), 4), lty=rep(c(1, 2), 4), ylab="No. Variables")
            }
            if(ms==300){
                boxplot(cbind(res$err.m[, 50], res$err.m2[,50], res$err.m[, 100], res$err.m2[,100], res$err.m[, 150], res$err.m2[, 150], res$err.m[,200], res$err.m2[, 200], res$err.m[,250], res$err.m2[,250], res$err.m[,300], res$err.m2[,300]), names=c("50", "50", "100", "100", "150", "150", "200", "200","250", "250", "300", "300"), col=rep(c("red", "blue"), 4), lty=rep(c(1, 2), 4), ylab="Misclassification Error")
                boxplot(cbind(res$nvar.m[, 50], res$nvar.m2[,50], res$nvar.m[, 100], res$nvar.m2[,100], res$nvar.m[, 150], res$nvar.m2[, 150], res$nvar.m[, 200], res$nvar.m2[, 200], res$nvar.m[,250], res$nvar.m2[,250], res$nvar.m[,300], res$nvar.m2[,300]), names=c("50", "50", "100", "100", "150", "150", "200", "200", "250", "250", "300", "300"), col=rep(c("red", "blue"), 4), lty=rep(c(1, 2), 4), ylab="No. Variables")
	    } 
            else if(ms==400){
                boxplot(cbind(res$err.m[, 100], res$err.m2[,100], res$err.m[,200], res$err.m2[, 200], res$err.m[, 300], res$err.m2[,300], res$err.m[, 400], res$err.m2[, 400]), col=rep(c("red", "blue"), 4), lty=rep(c(1, 2), 4),
                        names=c("B100", "TB100", "B200", "TB200", "B300", "TB300", "B400", "TB400"), ylab="Misclassification Error")
                                        #names=c("B50", "TB50", "B100", "TB100", "B150", "TB150", "B200", "TB200"), col=rep(c("red", "blue"), 4), lty=rep(c(1, 2), 4), ylab="Misclassification Error")
                boxplot(cbind(res$nvar.m[, 100], res$nvar.m2[,100], res$nvar.m[, 200], res$nvar.m2[,200], res$nvar.m[, 300], res$nvar.m2[, 300], res$nvar.m[, 400], res$nvar.m2[, 400]), 
                        names=c("B100", "TB100", "B200", "TB200", "B300", "TB300", "B400", "TB400"),
                        col=rep(c("red", "blue"), 4), lty=rep(c(1, 2), 4),
                        ylab="No. Variables")
            }
            else if(ms==600){
                boxplot(cbind(res$err.m[, 100], res$err.m2[,100], res$err.m[,200], res$err.m2[, 200], res$err.m[, 300], res$err.m2[,300], res$err.m[, 400], res$err.m2[, 400], res$err.m[,500], res$err.m2[,500], res$err.m[,600], res$err.m2[,600]), col=rep(c("red", "blue"), 6), lty=rep(c(1, 2), 6),
                        names=c("100", "100", "200", "200", "300", "300", "400", "400", "500", "500", "600", "600"), ylab="Misclassification Error")
                                        #names=c("B50", "TB50", "B100", "TB100", "B150", "TB150", "B200", "TB200"), col=rep(c("red", "blue"), 4), lty=rep(c(1, 2), 4), ylab="Misclassification Error")
                boxplot(cbind(res$nvar.m[, 100], res$nvar.m2[,100], res$nvar.m[, 200], res$nvar.m2[,200], res$nvar.m[, 300], res$nvar.m2[, 300], res$nvar.m[, 400], res$nvar.m2[, 400], res$nvar.m[,500], res$nvar.m2[,500], res$nvar.m[,600], res$nvar.m2[,600]), 
                        names=c("100", "100", "200", "200", "300", "300", "400", "400", "500", "500", "600", "600"), 
                        col=rep(c("red", "blue"), 6), lty=rep(c(1, 2), 6),
                        ylab="No. Variables")
            }
            dev.off()
        }
    }
    if(i==nrun){
        filename <- paste("ex", ex, "p", d, "con", k,"_rf", rf, "_bstfam", bstfam, "_twin", twin, "_n", ntr, "_rho",rho, learner, ".Rda", sep="")
        cat("saved RDa file name: ", filename, "\n")
	#save(file=filename, res)
    }
}
sessionInfo();
q("no")

